using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

using Microsoft.Kinect;
using System.IO;
using System.Threading;

namespace KinectSoundProcessor
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;


        // Game Data
        SpriteFont messageFont;
        Rectangle fullScreen;
        string message;
        KinectSensor myKinect;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // load the font
            messageFont = Content.Load<SpriteFont>("MessageFont");

            fullScreen = new Rectangle(0, 0, GraphicsDevice.Viewport.Width, GraphicsDevice.Viewport.Height);

            if (setupKinect())
            {
                message = "Microphone Ready";

                setupSoundDisplay();

                Thread audioThread = new Thread(new ThreadStart(pumpAudio));

                audioThread.Priority = ThreadPriority.Highest;

                audioThread.Start();
            }
        }

        protected bool setupKinect()
        {
            if (KinectSensor.KinectSensors.Count == 0)
            {
                message = "No Kinects detected";
                return false;
            }

            myKinect = KinectSensor.KinectSensors[0];

            try
            {
                myKinect.Start();
            }
            catch
            {
                message = "Kinect start failed";
                return false;
            }

            return true;
        }

        Texture2D soundTexture = null;
        byte[] soundTextureBytes;
        byte soundRed = 255;
        byte soundBlue = 255;
        byte soundGreen = 255;
        byte[] soundSampleBuffer;

        void setupSoundDisplay()
        {
            int width = GraphicsDevice.Viewport.Width;
            int height = GraphicsDevice.Viewport.Height;
            soundTextureBytes = new byte[width * height * 4];
            soundSampleBuffer = new byte[width * 2];
            // Create a new texture for the latest display
            soundTexture = new Texture2D(GraphicsDevice, width, height);
        }

        void updateSoundDisplay()
        {
            int width = GraphicsDevice.Viewport.Width;
            int height = GraphicsDevice.Viewport.Height;
            int halfHeight = height / 2;
            int rowPixelBytes = width * 4;

            // clear the screen from last time
            for (int i = 0; i < soundTextureBytes.Length; i++)
                soundTextureBytes[i] = 0;

            int samplePos = 0;
            // draw the scope trace on the screen
            for (int x = 0; x < width; x++)
            {
                // Compute the sound level
                short soundLevel = (short)(soundSampleBuffer[samplePos] | 
                                          (soundSampleBuffer[samplePos + 1] << 8));

                // move on to the next sound sample
                samplePos += 2;

                // Compute the position of this pixel
                int soundY = ((soundLevel * height) / 65535) + halfHeight;
                soundY = soundY * rowPixelBytes;
                int pixelPos = (x * 4) + soundY;

                // Set the color of the pixel
                soundTextureBytes[pixelPos] = soundRed;
                pixelPos++;
                soundTextureBytes[pixelPos] = soundBlue;
                pixelPos++;
                soundTextureBytes[pixelPos] = soundGreen;
                pixelPos++;
            }
            // Set the texture to the calculated oscilloscope display
            soundTexture.SetData(soundTextureBytes);
        }

        bool pumpingAudio = true;

        void pumpAudio()
        {
            Stream kinectAudioStream;
            // Start the source and connect the stream to it
            kinectAudioStream = myKinect.AudioSource.Start();

            // Create an output soundeffect to play the audio
            DynamicSoundEffectInstance soundEffectOutput;
            soundEffectOutput = new DynamicSoundEffectInstance(16000, AudioChannels.Mono);

            // Start the soundeffect playing
            soundEffectOutput.Play();

            // Repeatedly read the sound in and then play it
            while (pumpingAudio)
            {
                int count = kinectAudioStream.Read(soundSampleBuffer, 0, soundSampleBuffer.Length);
                soundEffectOutput.SubmitBuffer(soundSampleBuffer, 0, count);
            }

            myKinect.Stop();
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
            {
                pumpingAudio = false;
                this.Exit();
            }

            updateSoundDisplay();

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            spriteBatch.Begin();
            spriteBatch.Draw(soundTexture, fullScreen, Color.White);
            // Draw the score message in the middle of the screen
            float messageWidth =
                messageFont.MeasureString(message).X;

            float messageX =
                (GraphicsDevice.Viewport.Width - messageWidth) / 2;

            Vector2 messagePosition = new Vector2(messageX, 10);

            spriteBatch.DrawString(
                messageFont,        // font to use
                message,            // message string
                messagePosition,    // position on display
                Color.White);      // colour of text

            spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}
