﻿
// If SCREENSHOT_PAUSE is defined the program will stop updating the display
// after a set number of skeleton tracked events. 
// This is very useful if you want to take screenshots of the
// program output, and there is only one of you....


//#define SCREENSHOT_PAUSE

// If HIT_RADIUS is defined the program will draw a hit radius display around
// the mallet to show the active hit area.

//#define HIT_RADIUS

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using System.Threading;
using Microsoft.Kinect;

using System.Media;

using KinectUtils;

namespace BugSplat
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();
        }

#if SCREENSHOT_PAUSE

        // Used to pause the screen after a number of skeleton tracked events
        // This is so that I can take screenshots 
        // The trackcount and limit
        int trackCount = 0;
        int trackLimit = 200;

#endif

        #region Background Image Bitmap Creation

        BitmapSource gameImageBitmap;
        Byte[] gameImageBytes;

        /// <summary>
        /// Load the background image and create a bitmap for use when masking
        /// </summary>
        void setupGameImage()
        {
            gameImageBitmap = new BitmapImage(new Uri("BugsplatBackground.png", UriKind.RelativeOrAbsolute));

            gameImageBytes = new byte[gameImageBitmap.PixelWidth * gameImageBitmap.PixelHeight * 4];

            gameImageBitmap.CopyPixels(gameImageBytes, 1280, 0);
        }

        #endregion

        #region Depth Camera Background Isolation

        byte[] maskBytes = null;

        void createPlayerMaskBitmap(short[] depthValues, int width, int height)
        {
            /// Create a new mask buffer if this is the first time we have been called
            if (maskBytes == null)
                maskBytes = new byte[gameImageBytes.Length];

            // Copy a clean version of the game page into the mask
            Buffer.BlockCopy(gameImageBytes, 0, maskBytes, 0, gameImageBytes.Length);

            int gameImagePos;

            for (int depthPos = 0; depthPos < depthValues.Length; depthPos++)
            {
                int depthValue = depthValues[depthPos];

                int playerNo = depthValue & DepthImageFrame.PlayerIndexBitmask;
                depthValue = depthValue >> DepthImageFrame.PlayerIndexBitmaskWidth;

                if (playerNo != 0)
                {
                    // have a player - make them show through the mask

                    // find the X and Y positions of the depth point
                    int x = depthPos % width;
                    int y = depthPos / width;

                    // get the X and Y positions in the video image
                    ColorImagePoint playerPoint = myKinect.MapDepthToColorImagePoint(
                        DepthImageFormat.Resolution320x240Fps30,
                        x, y,
                        depthValues[depthPos],
                        ColorImageFormat.RgbResolution640x480Fps30);

                    // Divide the image position to match the mask
                    playerPoint.X /= 2;
                    playerPoint.Y /= 2;

                    // Make sure the point is on the screen
                    if (playerPoint.X < 0 || playerPoint.X >= 320 || playerPoint.Y < 0 || playerPoint.Y >= 240)
                        continue;

                    // Calculate the position in the mask bitmap
                    gameImagePos = (playerPoint.X + (playerPoint.Y * width)) * 4;

                    // Make the pixel transparent
                    maskBytes[gameImagePos] = 0;
                    gameImagePos++;
                    // Green
                    maskBytes[gameImagePos] = 0;
                    gameImagePos++;
                    // Red
                    maskBytes[gameImagePos] = 0;
                    gameImagePos++;
                    // transparency
                    maskBytes[gameImagePos] = 0;
                    gameImagePos++;
                }
            }
        }

        short[] depthData = null;
        int depthWidth;
        int depthHeight;
        WriteableBitmap depthImageBitmap = null;

        void processDepthFrame(DepthImageFrame depthFrame)
        {

#if SCREENSHOT_PAUSE
            if (trackCount == trackLimit) return;
#endif

            if (depthData == null)
                depthData = new short[depthFrame.Width * depthFrame.Height];

            depthFrame.CopyPixelDataTo(depthData);

            createPlayerMaskBitmap(depthData, depthFrame.Width, depthFrame.Height);
            depthWidth = depthFrame.Width;
            depthHeight = depthFrame.Height;
            Dispatcher.Invoke(new Action(() => drawDepthFrame()));
        }

        void drawDepthFrame()
        {
            if (maskBytes == null) return;

            // we now have a new array of color data

            if (depthImageBitmap == null)
            {
                this.depthImageBitmap = new WriteableBitmap(
                    depthWidth,
                    depthHeight,
                    96,  // DpiX
                    96,  // DpiY
                    PixelFormats.Bgra32,
                    null);
                gameImage.Source = depthImageBitmap;
            }

            this.depthImageBitmap.WritePixels(
                new Int32Rect(0, 0, depthWidth, depthHeight),
                maskBytes, // video data
                depthWidth * 4, // stride,
                0   // offset into the array - start at 0
                );
        }

        #endregion

        #region Player Video Display

        byte[] colorData = null;

        WriteableBitmap colorImage = null;
        int frameWidth;
        int frameHeight;

        private static readonly int Bgr32BytesPerPixel = (PixelFormats.Bgr32.BitsPerPixel + 7) / 8;

        void processColorFrame(ColorImageFrame colorFrame)
        {
#if SCREENSHOT_PAUSE
            if (trackCount == trackLimit) return;
#endif

            frameWidth = colorFrame.Width;
            frameHeight = colorFrame.Height;
            if (colorData == null)
                colorData = new byte[colorFrame.PixelDataLength];

            colorFrame.CopyPixelDataTo(colorData);
            Dispatcher.Invoke(new Action(() => drawColorFrame()));
        }

        void drawColorFrame()
        {
            if (colorData == null) return;

            if (colorImage == null)
            {
                this.colorImage = new WriteableBitmap(
                    frameWidth,
                    frameHeight,
                    96,  // DpiX
                    96,  // DpiY
                    PixelFormats.Bgr32,
                    null);
                kinectVideoImage.Source = colorImage;
            }

            this.colorImage.WritePixels(
                new Int32Rect(0, 0, frameWidth, frameHeight),
                this.colorData,
                frameWidth * Bgr32BytesPerPixel,
                0);
        }

        #endregion

        #region Skeleton and Mallet Management

        #region Kinect Skeleton Tracking

        Skeleton[] skeletons = null;
        string qualityMessage = "";

        void processSkeletonFrame(SkeletonFrame frame)
        {
#if SCREENSHOT_PAUSE
            if (trackCount == trackLimit) return;
#endif

            skeletons = new Skeleton[frame.SkeletonArrayLength];
            frame.CopySkeletonDataTo(skeletons);
            Dispatcher.Invoke(new Action(() => drawSkeleton()));
        }

        void drawSkeleton()
        {
            malletValid = false;

            if (skeletons == null) return;

            malletCanvas.Children.Clear();

            Brush brush = new SolidColorBrush(Colors.Red);

            qualityMessage = "";

            foreach (Skeleton skeleton in skeletons)
            {
                if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                {

#if SCREENSHOT_PAUSE
                    // got a tracked skeleton, update the counter
                    trackCount++;
#endif
                    if (skeleton.ClippedEdges == 0)
                    {
                        qualityMessage = "Good Quality";
                    }
                    else
                    {
                        if ((skeleton.ClippedEdges & FrameEdges.Bottom) != 0)
                            qualityMessage += "Move up\n";
                        if ((skeleton.ClippedEdges & FrameEdges.Top) != 0)
                            qualityMessage += "Move down\n";
                        if ((skeleton.ClippedEdges & FrameEdges.Right) != 0)
                            qualityMessage += "Move left\n";
                        if ((skeleton.ClippedEdges & FrameEdges.Left) != 0)
                            qualityMessage += "Move right\n";
                    }

                    // Uncomment if you want to draw the bones of the player
                    //drawSkeleton(skeleton);

                    updateMallet(skeleton.Joints[JointType.WristRight], skeleton.Joints[JointType.HandRight]);
                }
            }

            SkeletonQualityTextBlock.Text = qualityMessage;
        }

        #endregion

        #region Mallet drawing and update

        Brush malletHandleBrush = new SolidColorBrush(Colors.Black);
        Brush malletHeadBrush = new SolidColorBrush(Colors.Red);
        float malletLength = 100;
        float headLength = 50;

        System.Windows.Vector malletPosition;
        float malletHitRadius = 40;
        bool malletValid = false;

        void updateMallet(Joint j1, Joint j2)
        {
            if (j1.TrackingState != JointTrackingState.Tracked ||
                j2.TrackingState != JointTrackingState.Tracked)
                return;

            // Get the start and end positions of the mallet vector

            ColorImagePoint j1P = myKinect.MapSkeletonPointToColor(j1.Position, ColorImageFormat.RgbResolution640x480Fps30);
            ColorImagePoint j2P = myKinect.MapSkeletonPointToColor(j2.Position, ColorImageFormat.RgbResolution640x480Fps30);

            int dX = j2P.X - j1P.X;
            int dY = j2P.Y - j1P.Y;

            System.Windows.Vector malletDirection = new System.Windows.Vector(dX, dY);

            if (malletDirection.Length < 1) return;

            // Convert into a vector of length 1 unit
            malletDirection.Normalize();

            // now set the length of the mallet 
            System.Windows.Vector handleVector = malletDirection * malletLength;

            Line handleLine = new Line();

            handleLine.Stroke = malletHandleBrush;
            handleLine.StrokeThickness = 10;

            handleLine.X1 = j1P.X;
            handleLine.Y1 = j1P.Y;

            handleLine.X2 = j1P.X + handleVector.X;
            handleLine.Y2 = j1P.Y + handleVector.Y;

            malletCanvas.Children.Add(handleLine);

            Line headLine = new Line();

            headLine.Stroke = malletHeadBrush;
            headLine.StrokeThickness = 50;

            System.Windows.Vector headVector = malletDirection * headLength;

            headLine.X1 = handleLine.X2;
            headLine.Y1 = handleLine.Y2;

            headLine.X2 = handleLine.X2 + headVector.X;
            headLine.Y2 = handleLine.Y2 + headVector.Y;
            malletCanvas.Children.Add(headLine);

            malletPosition = new System.Windows.Vector(j1P.X, j1P.Y);

            malletPosition = malletPosition + (malletDirection * (malletLength + (headLength / 2)));

#if HIT_RADIUS

            // draw an ellipse around the mallet to show the hit radius

            Ellipse hitSpot = new Ellipse();
            hitSpot.Width = malletHitRadius * 2;
            hitSpot.Height = malletHitRadius * 2;
            hitSpot.Stroke = new SolidColorBrush(Colors.Yellow);
            malletCanvas.Children.Add(hitSpot);
            Canvas.SetLeft(hitSpot, malletPosition.X - malletHitRadius);
            Canvas.SetTop(hitSpot, malletPosition.Y - malletHitRadius);
#endif

            malletValid = true;
        }
        #endregion

        #region Skeleton Drawing


        Brush skeletonBrush = new SolidColorBrush(Colors.Red);

        void addLine(Joint j1, Joint j2)
        {
            Line boneLine = new Line();
            boneLine.Stroke = skeletonBrush;
            boneLine.StrokeThickness = 5;

            ColorImagePoint j1P = myKinect.MapSkeletonPointToColor(j1.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X1 = j1P.X;
            boneLine.Y1 = j1P.Y;

            ColorImagePoint j2P = myKinect.MapSkeletonPointToColor(j2.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X2 = j2P.X;
            boneLine.Y2 = j2P.Y;

            malletCanvas.Children.Add(boneLine);
        }

        void drawSkeleton(Skeleton skeleton)
        {
            // Spine
            addLine(skeleton.Joints[JointType.Head], skeleton.Joints[JointType.ShoulderCenter]);
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.Spine]);

            // Left leg
            addLine(skeleton.Joints[JointType.Spine], skeleton.Joints[JointType.HipCenter]);
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipLeft]);
            addLine(skeleton.Joints[JointType.HipLeft], skeleton.Joints[JointType.KneeLeft]);
            addLine(skeleton.Joints[JointType.KneeLeft], skeleton.Joints[JointType.AnkleLeft]);
            addLine(skeleton.Joints[JointType.AnkleLeft], skeleton.Joints[JointType.FootLeft]);

            // Right leg
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipRight]);
            addLine(skeleton.Joints[JointType.HipRight], skeleton.Joints[JointType.KneeRight]);
            addLine(skeleton.Joints[JointType.KneeRight], skeleton.Joints[JointType.AnkleRight]);
            addLine(skeleton.Joints[JointType.AnkleRight], skeleton.Joints[JointType.FootRight]);

            // Left arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderLeft]);
            addLine(skeleton.Joints[JointType.ShoulderLeft], skeleton.Joints[JointType.ElbowLeft]);
            addLine(skeleton.Joints[JointType.ElbowLeft], skeleton.Joints[JointType.WristLeft]);
            addLine(skeleton.Joints[JointType.WristLeft], skeleton.Joints[JointType.HandLeft]);

            // Right arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderRight]);
            addLine(skeleton.Joints[JointType.ShoulderRight], skeleton.Joints[JointType.ElbowRight]);
            addLine(skeleton.Joints[JointType.ElbowRight], skeleton.Joints[JointType.WristRight]);
            addLine(skeleton.Joints[JointType.WristRight], skeleton.Joints[JointType.HandRight]);

        }

        #endregion

        #endregion

        #region Kinect setup

        KinectManager myKinectMgr;
        KinectSensor myKinect;

        void setupKinect()
        {
            try
            {
                myKinectMgr = new KinectManager(
                    0, // Sensor number 0
                    ColorImageFormat.RgbResolution640x480Fps30, // Color image needed
                    DepthImageFormat.Resolution320x240Fps30,    // Depth image needed
                    true);                                      // Skeleton image needed

                myKinect = myKinectMgr.Kinect;

                myKinectMgr.DoProcessVideoFrame +=
                    new KinectManager.ProcessVideoFrame(processColorFrame);
                myKinectMgr.DoProcessDepthFrame +=
                    new KinectManager.ProcessDepthFrame(processDepthFrame);
                myKinectMgr.DoProcessSkeletons +=
                    new KinectManager.ProcessSkeletons(processSkeletonFrame);

                myKinectMgr.DoStatusChanged +=
                    new KinectManager.StatusChanged(updateStatus);

                myKinectMgr.Start();
            }
            catch
            {
                MessageBox.Show("Kinect initialise failed", "Bugsplat Game");
                Application.Current.Shutdown();
            }
        }

        #region Status Display

        string newStatus;

        void updateStatus(string status)
        {
            newStatus = status;
            Dispatcher.Invoke(new Action(() => updateMessage()));
        }

        void updateMessage()
        {
            KinectStatusTextBlock.Text = newStatus;
        }

        #endregion

        bool gameActive = true;

        #endregion

        #region Gameplay update

        private void startGameThread()
        {
            Thread game = new Thread(gameLoop);
            game.Start();
        }

        private void gameLoop()
        {
            while (gameActive)
            {
                Dispatcher.Invoke(new Action(() => updateGame()));
                Thread.Sleep(17);
            }
        }

        double bugY = 0;
        double bugX = 0;

        double bugYSpeed = 1;

        Random rand = new Random(1);

        int score = 0;

        SoundPlayer scorePlayer;

        private void updateGame()
        {
            Canvas.SetTop(bugImage, bugY);
            bugY = bugY + bugYSpeed;

            // check for collisions

            if (malletValid)
            {
                double bugCenterX = bugX + (bugImage.Width / 2);
                double bugCenterY = bugY + (bugImage.Height / 2);

                System.Windows.Vector hitVector =
                    new System.Windows.Vector(malletPosition.X - bugCenterX, malletPosition.Y - bugCenterY);
                if (hitVector.Length < malletHitRadius)
                {
                    score++;
                    ScoreTextBlock.Text = score.ToString();
                    scorePlayer.Play();
                    bugX = rand.Next(0, (int)gameCanvas.Width - (int)bugImage.Width);
                    Canvas.SetLeft(bugImage, bugX);
                    bugY = -bugImage.Height;
                }
            }

            if (bugY > gameCanvas.Height)
            {
                bugX = rand.Next(0, (int)gameCanvas.Width - (int)bugImage.Width);
                Canvas.SetLeft(bugImage, bugX);
                bugY = -bugImage.Height;
            }
        }

        #endregion

        #region Window Event Handlers

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            setupGameImage();
            setupKinect();
            startGameThread();
            scorePlayer = new SoundPlayer("sounds/ding.wav");
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            gameActive = false;
            if (myKinectMgr != null)
                myKinectMgr.Stop();
        }

        #endregion
    }
}
